import yaml
from yaml.loader import SafeLoader

from apistar.formatters.nodes import DictNode, ListNode, LiteralNode


def parse_yaml(content):
    class CustomLoader(SafeLoader):
        pass

    def construct_mapping(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        mapping = loader.construct_mapping(node)
        plain_mapping = {}
        for key, value in mapping.items():
            value.key_start = key.start
            value.key_end = key.end
            plain_mapping[key.value] = value
        return DictNode(plain_mapping, start, end)

    def construct_sequence(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_sequence(node)
        return ListNode(value, start, end)

    def construct_scalar(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_scalar(node)
        return LiteralNode(value, start, end)

    def construct_int(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_yaml_int(node)
        return LiteralNode(value, start, end)

    def construct_float(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_yaml_float(node)
        return LiteralNode(value, start, end)

    def construct_bool(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_yaml_bool(node)
        return LiteralNode(value, start, end)

    def construct_null(loader, node):
        start = node.start_mark.index
        end = node.end_mark.index
        value = loader.construct_yaml_null(node)
        return LiteralNode(value, start, end)

    CustomLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)

    CustomLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_SEQUENCE_TAG,
        construct_sequence)

    CustomLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_SCALAR_TAG,
        construct_scalar)

    CustomLoader.add_constructor(
        'tag:yaml.org,2002:int',
        construct_int)

    CustomLoader.add_constructor(
        'tag:yaml.org,2002:float',
        construct_float)

    CustomLoader.add_constructor(
        'tag:yaml.org,2002:bool',
        construct_bool)

    CustomLoader.add_constructor(
        'tag:yaml.org,2002:null',
        construct_null)

    return yaml.load(content, CustomLoader)



class LiteralNode:
    def __init__(self, value, start, end):
        self.value = value
        self.start = start
        self.end = end
        self.key_start = None
        self.key_end = None

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return repr(self.value)

    def lookup(self, prefix):
        if not prefix:
            return self
        return self[prefix[0]].lookup(prefix[1:])


class DictNode(dict):
    def __init__(self, value, start, end):
        super().__init__(value)
        self.start = start
        self.end = end
        self.key_start = None
        self.key_end = None

    def __repr__(self):
        return repr(dict(self))

    def lookup(self, prefix):
        if not prefix:
            return self
        return self[prefix[0]].lookup(prefix[1:])


class ListNode(list):
    def __init__(self, value, start, end):
        super().__init__(value)
        self.start = start
        self.end = end
        self.key_start = None
        self.key_end = None

    def __repr__(self):
        return repr(list(self))

    def lookup(self, prefix):
        if not prefix:
            return self
        return self[prefix[0]].lookup(prefix[1:])



from apistar.server.app import App, ASyncApp
from apistar.server.components import Component
from apistar.server.core import Include, Route

__all__ = ['App', 'ASyncApp', 'Component', 'Route', 'Include']



import json

from apistar import codecs
from apistar.types import Type


class _CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Type):
            return dict(obj)
        return json.JSONEncoder.default(self, obj)


def encode_json(data, indent=False):
    kwargs = {
        'ensure_ascii': False,
        'allow_nan': False,
        'cls': _CustomEncoder
    }
    if indent:
        kwargs.update({
            'indent': None,
            'separators': (',', ':')
        })
    else:
        kwargs.update({
            'indent': 4,
            'separators': (',', ': ')
        })
    return json.dumps(data, **kwargs).encode('utf-8')


def encode_jsonschema(validator, to_data_structure=False):
    codec = codecs.JSONSchemaCodec()
    return codec.encode(validator, to_data_structure=to_data_structure)
